#include <msp/datafile/collection.h>
#include <msp/gl/extensions/arb_texture_cube_map.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include <msp/io/memory.h>
#include <msp/strings/format.h>
#include "bindable.h"
#include "error.h"
#include "pixelstore.h"
#include "texturecube.h"

using namespace std;

namespace Msp {
namespace GL {

TextureCubeFace TextureCube::face_order[6] =
{
	POSITIVE_X,
	NEGATIVE_X,
	POSITIVE_Y,
	NEGATIVE_Y,
	POSITIVE_Z,
	NEGATIVE_Z
};

Vector3 TextureCube::directions[6] =
{
	Vector3(1, 0, 0),
	Vector3(-1, 0, 0),
	Vector3(0, 1, 0),
	Vector3(0, -1, 0),
	Vector3(0, 0, 1),
	Vector3(0, 0, -1)
};

unsigned TextureCube::orientations[12] =
{
	5, 3,
	4, 3,
	0, 4,
	0, 5,
	0, 3,
	1, 3
};

TextureCube::TextureCube():
	Texture(GL_TEXTURE_CUBE_MAP),
	size(0),
	allocated(0)
{
	static Require _req(ARB_texture_cube_map);
}

void TextureCube::storage(PixelFormat fmt, unsigned sz, unsigned lv)
{
	if(size>0)
		throw invalid_operation("TextureCube::storage");
	if(sz==0)
		throw invalid_argument("TextureCube::storage");

	set_format(fmt);
	size = sz;
	levels = get_n_levels();
	if(lv>0)
		levels = min(levels, lv);
}

void TextureCube::allocate(unsigned level)
{
	if(size==0)
		throw invalid_operation("TextureCube::allocate");
	if(level>=levels)
		throw invalid_argument("TextureCube::allocate");
	if(allocated&(64<<level))
		return;

	if(ARB_texture_storage)
	{
		BindRestore _bind(this);
		glTexStorage2D(target, levels, storage_fmt, size, size);
		apply_swizzle();
		allocated |= (1<<levels)-1;
	}
	else
	{
		for(unsigned i=0; i<6; ++i)
			image(enumerate_faces(i), level, 0);
	}
}

void TextureCube::image(TextureCubeFace face, unsigned level, const void *data)
{
	if(size==0)
		throw invalid_operation("TextureCube::image");

	unsigned s = get_level_size(level);
	if(s==0)
		throw out_of_range("TextureCube::image");

	if(ARB_texture_storage)
		return sub_image(face, level, 0, 0, s, s, data);

	BindRestore _bind(this);

	if(!allocated)
	{
		glTexParameteri(target, GL_TEXTURE_MAX_LEVEL, levels-1);
		apply_swizzle();
	}

	PixelComponents comp = get_components(storage_fmt);
	DataType type = get_component_type(storage_fmt);
	glTexImage2D(face, level, storage_fmt, s, s, 0, comp, type, data);

	if(level==0)
	{
		allocated |= 1<<get_face_index(face);
		if((allocated&63)==63)
		{
			allocated |= 64;
			if(auto_gen_mipmap)
			{
				generate_mipmap();
				allocated |= (64<<levels)-1;
			}
		}
	}
	else if(!(allocated&(64<<level)))
	{
		for(unsigned i=0; i<6; ++i)
			if(enumerate_faces(i)!=face)
				glTexImage2D(enumerate_faces(i), level, storage_fmt, s, s, 0, comp, type, 0);

		allocated |= 64<<level;
	}
}

void TextureCube::image(TextureCubeFace face, unsigned level, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("TextureCube::image");
	image(face, level, data);
}

void TextureCube::sub_image(TextureCubeFace face, unsigned level, int x, int y, unsigned wd, unsigned ht, const void *data)
{
	if(size==0)
		throw invalid_operation("TextureCube::sub_image");

	BindRestore _bind(this);
	allocate(level);

	PixelComponents comp = get_components(storage_fmt);
	DataType type = get_component_type(storage_fmt);
	glTexSubImage2D(face, level, x, y, wd, ht, comp, type, data);

	if(auto_gen_mipmap && level==0)
		generate_mipmap();
}

void TextureCube::sub_image(TextureCubeFace face, unsigned level, int x, int y, unsigned wd, unsigned ht, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("TextureCube::subimage");
	sub_image(face, level, x, y, wd, ht, data);
}

void TextureCube::image(TextureCubeFace face, const Graphics::Image &img)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();
	PixelFormat fmt = pixelformat_from_image(img);
	if(size==0)
	{
		if(w!=h)
			throw incompatible_data("TextureCube::image");

		storage(make_pixelformat(get_components(fmt), get_component_type(fmt), use_srgb_format), w);
	}
	else if(w!=size || h!=size)
		throw incompatible_data("TextureCube::image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(face, 0, img.get_pixels());
}

void TextureCube::image(TextureCubeFace face, const Graphics::Image &img, bool)
{
	image(face, img);
}

void TextureCube::image(const Graphics::Image &img, unsigned lv)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();

	if(h!=w*6)
		throw incompatible_data("TextureCube::image");
	h /= 6;

	PixelFormat fmt = pixelformat_from_image(img);
	if(size==0)
		storage(make_pixelformat(get_components(fmt), get_component_type(fmt), use_srgb_format), w, lv);
	else if(w!=size || h!=size)
		throw incompatible_data("TextureCube::image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	const char *pixels = reinterpret_cast<const char *>(img.get_pixels());
	unsigned face_size = img.get_stride()*size;
	for(unsigned i=0; i<6; ++i)
		image(enumerate_faces(i), 0, pixels+i*face_size);
}

unsigned TextureCube::get_n_levels() const
{
	unsigned n = 0;
	for(unsigned s=size; s; s>>=1, ++n) ;
	return n;
}

unsigned TextureCube::get_level_size(unsigned level) const
{
	return size>>level;
}

TextureCubeFace TextureCube::enumerate_faces(unsigned i)
{
	if(i>=6)
		throw out_of_range("TextureCube::enumerate_faces");
	return face_order[i];
}

unsigned TextureCube::get_face_index(TextureCubeFace face)
{
	switch(face)
	{
	case POSITIVE_X: return 0;
	case NEGATIVE_X: return 1;
	case POSITIVE_Y: return 2;
	case NEGATIVE_Y: return 3;
	case POSITIVE_Z: return 4;
	case NEGATIVE_Z: return 5;
	default: throw invalid_argument("TextureCube::get_face_index");
	}
}

const Vector3 &TextureCube::get_face_direction(TextureCubeFace face)
{
	return directions[get_face_index(face)];
}

const Vector3 &TextureCube::get_s_direction(TextureCubeFace face)
{
	return directions[orientations[get_face_index(face)*2]];
}

const Vector3 &TextureCube::get_t_direction(TextureCubeFace face)
{
	return directions[orientations[get_face_index(face)*2+1]];
}

Vector3 TextureCube::get_texel_direction(TextureCubeFace face, unsigned u, unsigned v)
{
	float s = (u+0.5f)*2.0f/size-1.0f;
	float t = (v+0.5f)*2.0f/size-1.0f;
	const Vector3 &fv = get_face_direction(face);
	const Vector3 &sv = get_s_direction(face);
	const Vector3 &tv = get_t_direction(face);
	return fv+s*sv+t*tv;
}

UInt64 TextureCube::get_data_size() const
{
	return id ? size*size*6*get_pixel_size(storage_fmt) : 0;
}


TextureCube::Loader::Loader(TextureCube &t):
	DataFile::DerivedObjectLoader<TextureCube, Texture::Loader>(t)
{
	init();
}

TextureCube::Loader::Loader(TextureCube &t, Collection &c):
	DataFile::DerivedObjectLoader<TextureCube, Texture::Loader>(t, c)
{
	init();
}

void TextureCube::Loader::init()
{
	add("external_image", &Loader::external_image);
	add("image_data", &Loader::image_data);
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
	add("storage", &Loader::storage_levels);
}

void TextureCube::Loader::external_image(TextureCubeFace face, const string &fn)
{
	Graphics::Image img;
	RefPtr<IO::Seekable> io = get_collection().open_raw(fn);
	img.load_io(*io);

	obj.image(face, img);
}

void TextureCube::Loader::image_data(TextureCubeFace face, const string &data)
{
	Graphics::Image img;
	IO::Memory mem(data.data(), data.size());
	img.load_io(mem);

	obj.image(face, img);
}

void TextureCube::Loader::raw_data(TextureCubeFace face, const string &data)
{
	obj.image(face, 0, data.data());
}

void TextureCube::Loader::storage(PixelFormat fmt, unsigned s)
{
	obj.storage(fmt, s);
}

void TextureCube::Loader::storage_levels(PixelFormat fmt, unsigned s, unsigned l)
{
	obj.storage(fmt, s, l);
}


void operator>>(const LexicalConverter &conv, TextureCubeFace &face)
{
	const string &str = conv.get();
	if(str=="POSITIVE_X")
		face = POSITIVE_X;
	else if(str=="NEGATIVE_X")
		face = NEGATIVE_X;
	else if(str=="POSITIVE_Y")
		face = POSITIVE_Y;
	else if(str=="NEGATIVE_Y")
		face = NEGATIVE_Y;
	else if(str=="POSITIVE_Z")
		face = POSITIVE_Z;
	else if(str=="NEGATIVE_Z")
		face = NEGATIVE_Z;
	else
		throw lexical_error(format("conversion of '%s' to TextureCubeFace", str));
}

} // namespace GL
} // namespace Msp
